@@ -49,6 +49,8 @@ class Handler(Base):
         25: LabelsList,
         26: LabelInterest,
         27: LabelIgnore,
+        28: GetHistory,
+        29: History,
     })
 
     def client(self, master):
@@ -91,7 +93,9 @@ class Handler(Base):
 
             while msg:
                 try:
-                    while msg := self.c_recv():
+                    #while msg := self.c_recv():
+                    msg = self.c_recv()
+                    while msg:
                         reply = None
 
                         if isinstance(msg, Pong):
@@ -123,6 +127,8 @@ class Handler(Base):
                                 msg = None
                                 break
 
+                        msg = self.c_recv()
+
                 except socket.timeout as e:
                     if ping: raise e
                     ping = Ping()
@@ -137,7 +143,9 @@ class Handler(Base):
         self.peer_label = None
 
         try:
-            while msg := self.s_recv():
+            #while msg := self.s_recv():
+            msg = self.s_recv()
+            while msg:
                 reply = None
 
                 if isinstance(msg, Ping):
@@ -176,6 +184,9 @@ class Handler(Base):
                 elif isinstance(msg, LabelIgnore):
                     reply = self.label_ignore(master, msg)
 
+                elif isinstance(msg, GetHistory):
+                    reply = self.get_history(master, msg)
+
                 else:
                     raise ValueError('unhandled message')
 
@@ -185,6 +196,8 @@ class Handler(Base):
                     if isinstance(reply, Denial):
                         break
 
+                msg = self.s_recv()
+
         finally:
             if self.node_uuid:
                 del master.nodes_by_uuid[self.node_uuid]
@@ -462,7 +475,13 @@ class Handler(Base):
                 signed['address'].encode(crypto.ENCODING))
             host = msg.address[:msg.address.index(':')]
             port = int(msg.address[msg.address.index(':')+1:])
-            data = PeerOn._to_sign_(crypto, msg.uuid, msg.label, host, port)
+            alias = None
+            if ':' in msg.key[msg.key.index(':')+1:]:
+                alias = msg.key[msg.key.index(':')+1:]
+                alias = alias[alias.index(':')+1:]
+            port = int(port)
+            data = PeerOn._to_sign_(crypto, msg.uuid, msg.label, host, port,
+                                    alias)
 
             if not crypto.verify(pub_key, data, msg.signature):
                 return SignatureFailure(msg.uuid)
@@ -564,6 +583,7 @@ class Handler(Base):
 
         assert ':' not in self.addr[0]
         key = f'{self.addr[0]}:{self.addr[1]}'
+        if msg.alias: key += f':{msg.alias}'
 
         assert ':' not in msg.host
         addr = f'{msg.host}:{msg.port}'
@@ -611,3 +631,38 @@ class Handler(Base):
 
         master.remove_label_interest(msg.label, self)
 
+    def get_history(self, master, msg):
+        assert isinstance(msg, GetHistory)
+
+        try:
+            label_id = master.labels_by_id.inverse[msg.label]
+
+        except KeyError:
+            return LabelNotFound(msg.label)
+
+        signed_list = master.ldb.find_signed_by_label(
+            label_id, msg.start + msg.count)
+
+        items = [ ]
+        idx = 0
+        end = msg.start + msg.count - 1
+
+        for signed in signed_list:
+            if idx >= msg.start:
+                items.append(Signed(
+                    signed['version'],
+                    UUID(signed['uuid']),
+                    msg.label,
+                    signed['name'],
+                    signed['key'],
+                    signed['address'],
+                    signed['signature'],
+                    UUID(signed['node_uuid']),
+                    signed['created']
+                ))
+
+            idx += 1
+            if idx > end: break
+
+        return History(msg.label, msg.start, items)
+
